/*
 * Copyright (c) 2001,2003,2004 David H. Hovemeyer <daveho@cs.umd.edu>
 * Copyright (c) 2003,2013,2014 Jeffrey K. Hollingsworth <hollings@cs.umd.edu>
 * Copyright (c) 2015 Neil Spring <nspring@cs.umd.edu>
 * Copyright (c) 2024 Leo Liu <leoliu@cs.umd.edu>
 *
 * All rights reserved.
 *
 * This code may not be resdistributed without the permission of the copyright holders.
 * Any student solutions using any of this code base constitute derviced work and may
 * not be redistributed in any form.  This includes (but is not limited to) posting on
 * public forums or web sites, providing copies to (past, present, or future) students
 * enrolled in similar operating systems courses the University of Maryland's CMSC412 course.
 */

#ifndef LFS_H
#define LFS_H

#include <geekos/bufcache.h>

/* this is a simplified version of log structured file system,
 * the idea is only allowing limited number of files,
 * using only block 1 as check point region
 */

#define LFS_MAGIC 0x4C474653
#define SECTOR_SIZE 512
#define ROOT_INODE 1
#define LFS_BOOT_RECORD_OFFSET 482

/* in unit of blocks, segment size set to 8k */
/* 2 pages */
#define SEGMENT_SIZE 16
#define SEGMENT_SIZE_BYTE (SEGMENT_SIZE * SECTOR_SIZE)

#define DATA_BLOCKS_PER_INODE 10
#define IMAP_OFFSET 1

#define IMAP_PER_BLOCK (SECTOR_SIZE / 4)
#define NUM_IMAPS 8

#define DIRECT_BLOCK 0
#define INDIRECT_BLOCK 1

#define LFS_DIRECTORY 1
#define LFS_FILE 2
#define MAX_INODES 1024         /* 128 imap entries in one block, using 8 blocks */

#define INODE_PER_BLCOK (SECTOR_SIZE/4)

typedef unsigned int lfs_inodenum;      /* For inode numbers */

/* last 9 bits offset in disk, others block number */
typedef unsigned int Address;

#define GET_ADDRESS(block_num, offset) ((block_num << 9) | offset)
#define GET_BLOCK_NUM(address) (address >> 9)
#define GET_OFFSET(address) (address & (0x1ff))

typedef struct lfs_superblock {
    unsigned int lfs_magic;     /* 0x4C474653, lgfs */
    unsigned int lfs_version;   /* 0x00000001 */
    unsigned int block_size;
    unsigned int max_check_point_blocks;        /* should only be 1 for now */
    unsigned int blocks_per_disk;
} Lfs_superblock;

typedef struct data_block {
    Address address;
    unsigned int size;
    unsigned char type;         /* direct 0 or indirect 1 */
} Data_block;

typedef struct lfs_inode {
    unsigned int size;
    unsigned char type;         /* directory = 1, file = 2 */
    unsigned char reference_count;      /* 1? */
    unsigned short mode;        /* permission */
    Data_block data[DATA_BLOCKS_PER_INODE];
} Lfs_inode;

// typedef struct imap {
//   Address inodes[INODE_PER_IMAP];
// } Imap;

/* exactly one block for check_point_region */
/* simplified lfs, we will not do check point region */
typedef struct check_point_region {
    Address imaps[IMAP_PER_BLOCK];
} CR;


/* note that a lfs_dirent is a variably sized structure */
/* using the same struct as gfs3_dirent... do I need a new one? */
typedef struct lfs_dirent {
    lfs_inodenum inum;          /* *must* be word-aligned */
    unsigned char entry_length; /* after the inum.  i.e., the min entry length is 4.  max is 252, for alignment */
    unsigned char name_length;  /* may be < entry length - 2. */
    char name[2];               /* where overrun provides the rest of the name; NOT (necessarily) null terminated. */
} Lfs_dirent;

typedef struct lfs_segment_summary {
    unsigned int summary_checksum;
    unsigned int data_checksum;
    Address next_segment;
    unsigned int create_time_high;      /* unix timestamp */
    unsigned int create_time_low;       /* unix timestamp */
    unsigned int size;
    unsigned int num_inodes;
    /* TODO: there should be more... */
    /* 24 bytes for now */

} Lfs_segment_summary;

typedef struct lfs_segment {
    Address address;
    void *segment;              /* pointer to memory  with segment summary at beginning */
    unsigned size;
} Lfs_segment;


void Init_LFS();

#endif /* LFS_H */
